from __future__ import annotations

__ZERO__ = b"MHp2Zg=="
# noinspection SpellCheckingInspection
__MOMONGA__ = b"NTh3Yw=="
__LIGHT__ = b"|OW5hMQ=="
__SEPHIROTH__ = b"ZzRlbA=="
ANIME = b"|".join([__SEPHIROTH__, __MOMONGA__, __ZERO__, __LIGHT__])

__all__ = ("ANIME",)
